@@ -3,35 +3,33 @@ Train a new model.
 """
 
 from __future__ import annotations
-import torch
-import torch.nn as nn
-import torch.nn.functional as F
-from torch.autograd import Variable
-from torch.utils.data import IterableDataset, DataLoader
-from sklearn.metrics import average_precision_score as average_precision
-from tqdm import tqdm
-from typing import Callable, NamedTuple, Optional
 
-import sys
 import argparse
+import sys
+from collections.abc import Callable
+from typing import NamedTuple
+
 import h5py
-import subprocess as sp
 import numpy as np
 import pandas as pd
-import gzip as gz
+import torch
+import torch.nn.functional as F
+from sklearn.metrics import average_precision_score as average_precision
+from torch.autograd import Variable
+from tqdm import tqdm
 
 from .. import __version__
-from ..alphabets import Uniprot21
+from ..fasta import parse_dict
+from ..foldseek import fold_vocab, get_foldseek_onehot
 from ..glider import glide_compute_map, glider_score
+from ..models.contact import ContactCNN
+from ..models.embedding import FullyConnectedEmbed
+from ..models.interaction import ModelInteraction
 from ..utils import (
     PairedDataset,
     collate_paired_sequences,
     log,
-    load_hdf5_parallel,
 )
-from ..models.embedding import FullyConnectedEmbed
-from ..models.contact import ContactCNN
-from ..models.interaction import ModelInteraction
 
 
 class TrainArguments(NamedTuple):
@@ -58,10 +56,10 @@ class TrainArguments(NamedTuple):
     run_tt: bool
     glider_weight: float
     glider_thresh: float
-    outfile: Optional[str]
-    save_prefix: Optional[str]
-    checkpoint: Optional[str]
-    seed: Optional[int]
+    outfile: str | None
+    save_prefix: str | None
+    checkpoint: str | None
+    seed: int | None
     func: Callable[[TrainArguments], None]
 
 
@@ -78,11 +76,10 @@ def add_args(parser):
     inter_grp = parser.add_argument_group("Interaction Module")
     train_grp = parser.add_argument_group("Training")
     misc_grp = parser.add_argument_group("Output and Device")
+    foldseek_grp = parser.add_argument_group("Foldseek related commands")
 
     # Data
-    data_grp.add_argument(
-        "--train", required=True, help="list of training pairs"
-    )
+    data_grp.add_argument("--train", required=True, help="list of training pairs")
     data_grp.add_argument(
         "--test", required=True, help="list of validation/testing pairs"
     )
@@ -211,24 +208,45 @@ def add_args(parser):
     )
 
     # Output
-    misc_grp.add_argument(
-        "-o", "--outfile", help="output file path (default: stdout)"
-    )
-    misc_grp.add_argument(
-        "--save-prefix", help="path prefix for saving models"
-    )
+    misc_grp.add_argument("-o", "--outfile", help="output file path (default: stdout)")
+    misc_grp.add_argument("--save-prefix", help="path prefix for saving models")
     misc_grp.add_argument(
         "-d", "--device", type=int, default=-1, help="compute device to use"
     )
-    misc_grp.add_argument(
-        "--checkpoint", help="checkpoint model to start training from"
-    )
+    misc_grp.add_argument("--checkpoint", help="checkpoint model to start training from")
     misc_grp.add_argument("--seed", help="Set random seed", type=int)
 
+    ## Foldseek arguments
+    foldseek_grp.add_argument(
+        "--allow_foldseek",
+        default=False,
+        action="store_true",
+        help="If set to true, adds the foldseek one-hot representation",
+    )
+    foldseek_grp.add_argument(
+        "--foldseek_fasta",
+        help="foldseek fasta file containing the foldseek representation",
+    )
+    # foldseek_grp.add_argument(
+    #     "--add_foldseek_after_projection", default = False, action = "store_true", help = "If set to true, adds the fold seek embedding after the projection layer"
+    # )
+
     return parser
 
 
-def predict_cmap_interaction(model, n0, n1, tensors, use_cuda):
+def predict_cmap_interaction(
+    model,
+    n0,
+    n1,
+    tensors,
+    use_cuda,
+    ### Foldseek added here
+    allow_foldseek=False,
+    fold_record=None,
+    fold_vocab=None,
+    add_first=True,
+    ###
+):
     """
     Predict whether a list of protein pairs will interact, as well as their contact map.
 
@@ -249,12 +267,33 @@ def predict_cmap_interaction(model, n0, n1, tensors, use_cuda):
     p_hat = []
     c_map_mag = []
     for i in range(b):
-        z_a = tensors[n0[i]]
+        z_a = tensors[n0[i]]  # 1 x seqlen x dim
         z_b = tensors[n1[i]]
         if use_cuda:
             z_a = z_a.cuda()
             z_b = z_b.cuda()
-        cm, ph = model.map_predict(z_a, z_b)
+        if allow_foldseek:
+            assert fold_record is not None and fold_vocab is not None
+            f_a = get_foldseek_onehot(
+                n0[i], z_a.shape[1], fold_record, fold_vocab
+            ).unsqueeze(0)  # seqlen x vocabsize
+            f_b = get_foldseek_onehot(
+                n1[i], z_b.shape[1], fold_record, fold_vocab
+            ).unsqueeze(0)
+
+            ## check if cuda
+            if use_cuda:
+                f_a = f_a.cuda()
+                f_b = f_b.cuda()
+
+            if add_first:
+                z_a = torch.concat([z_a, f_a], dim=2)
+                z_b = torch.concat([z_b, f_b], dim=2)
+
+        if allow_foldseek and (not add_first):
+            cm, ph = model.map_predict(z_a, z_b, True, f_a, f_b)
+        else:
+            cm, ph = model.map_predict(z_a, z_b)
         p_hat.append(ph)
         c_map_mag.append(torch.mean(cm))
     p_hat = torch.stack(p_hat, 0)
@@ -262,7 +301,19 @@ def predict_cmap_interaction(model, n0, n1, tensors, use_cuda):
     return c_map_mag, p_hat
 
 
-def predict_interaction(model, n0, n1, tensors, use_cuda):
+def predict_interaction(
+    model,
+    n0,
+    n1,
+    tensors,
+    use_cuda,
+    ### Foldseek added here
+    allow_foldseek=False,
+    fold_record=None,
+    fold_vocab=None,
+    add_first=True,
+    ###
+):
     """
     Predict whether a list of protein pairs will interact.
 
@@ -277,7 +328,17 @@ def predict_interaction(model, n0, n1, tensors, use_cuda):
     :param use_cuda: Whether to use GPU
     :type use_cuda: bool
     """
-    _, p_hat = predict_cmap_interaction(model, n0, n1, tensors, use_cuda)
+    _, p_hat = predict_cmap_interaction(
+        model,
+        n0,
+        n1,
+        tensors,
+        use_cuda,
+        allow_foldseek,
+        fold_record,
+        fold_vocab,
+        add_first,
+    )
     return p_hat
 
 
@@ -293,6 +354,12 @@ def interaction_grad(
     glider_map=None,
     glider_mat=None,
     use_cuda=True,
+    ### Foldseek added here
+    allow_foldseek=False,
+    fold_record=None,
+    fold_vocab=None,
+    add_first=True,
+    ###
 ):
     """
     Compute gradient and backpropagate loss for a batch.
@@ -325,7 +392,15 @@ def interaction_grad(
     """
 
     c_map_mag, p_hat = predict_cmap_interaction(
-        model, n0, n1, tensors, use_cuda
+        model,
+        n0,
+        n1,
+        tensors,
+        use_cuda,
+        allow_foldseek,
+        fold_record,
+        fold_vocab,
+        add_first,
     )
 
     if use_cuda:
@@ -349,9 +424,7 @@ def interaction_grad(
             g_score = g_score.cuda()
 
         glider_loss = F.binary_cross_entropy(p_hat.float(), g_score.float())
-        accuracy_loss = (glider_weight * glider_loss) + (
-            (1 - glider_weight) * bce_loss
-        )
+        accuracy_loss = (glider_weight * glider_loss) + ((1 - glider_weight) * bce_loss)
     else:
         accuracy_loss = bce_loss
 
@@ -381,7 +454,18 @@ def interaction_grad(
     return loss, correct, mse, b
 
 
-def interaction_eval(model, test_iterator, tensors, use_cuda):
+def interaction_eval(
+    model,
+    test_iterator,
+    tensors,
+    use_cuda,
+    ### Foldseek added here
+    allow_foldseek=False,
+    fold_record=None,
+    fold_vocab=None,
+    add_first=True,
+    ###
+):
     """
     Evaluate test data set performance.
 
@@ -401,7 +485,19 @@ def interaction_eval(model, test_iterator, tensors, use_cuda):
     true_y = []
 
     for n0, n1, y in test_iterator:
-        p_hat.append(predict_interaction(model, n0, n1, tensors, use_cuda))
+        p_hat.append(
+            predict_interaction(
+                model,
+                n0,
+                n1,
+                tensors,
+                use_cuda,
+                allow_foldseek,
+                fold_record,
+                fold_vocab,
+                add_first,
+            )
+        )
         true_y.append(y)
 
     y = torch.cat(true_y, 0)
@@ -438,7 +534,6 @@ def interaction_eval(model, test_iterator, tensors, use_cuda):
 
 def train_model(args, output):
     # Create data sets
-
     batch_size = args.batch_size
     use_cuda = (args.device > -1) and torch.cuda.is_available()
     train_fi = args.train
@@ -446,7 +541,20 @@ def train_model(args, output):
     no_augment = args.no_augment
 
     embedding_h5 = args.embedding
-    # h5fi = h5py.File(embedding_h5, "r")
+
+    ########## Foldseek code #########################3
+    allow_foldseek = args.allow_foldseek
+    fold_fasta_file = args.foldseek_fasta
+    # fold_vocab_file = args.foldseek_vocab
+    add_first = False
+    fold_record = {}
+    # fold_vocab = None
+    if allow_foldseek:
+        assert fold_fasta_file is not None
+        fold_fasta = parse_dict(fold_fasta_file)
+        for rec_k, rec_v in fold_fasta.items():
+            fold_record[rec_k] = rec_v
+    ##################################################
 
     train_df = pd.read_csv(train_fi, sep="\t", header=None)
     train_df.columns = ["prot1", "prot2", "label"]
@@ -456,12 +564,12 @@ def train_model(args, output):
         train_p2 = train_df["prot2"]
         train_y = torch.from_numpy(train_df["label"].values)
     else:
-        train_p1 = pd.concat(
-            (train_df["prot1"], train_df["prot2"]), axis=0
-        ).reset_index(drop=True)
-        train_p2 = pd.concat(
-            (train_df["prot2"], train_df["prot1"]), axis=0
-        ).reset_index(drop=True)
+        train_p1 = pd.concat((train_df["prot1"], train_df["prot2"]), axis=0).reset_index(
+            drop=True
+        )
+        train_p2 = pd.concat((train_df["prot2"], train_df["prot1"]), axis=0).reset_index(
+            drop=True
+        )
         train_y = torch.from_numpy(
             pd.concat((train_df["label"], train_df["label"])).values
         )
@@ -495,11 +603,13 @@ def train_model(args, output):
     log("Loading embeddings...", file=output)
     output.flush()
 
-    # embeddings = {}
     all_proteins = set(train_p1).union(train_p2).union(test_p1).union(test_p2)
-    # for prot_name in tqdm(all_proteins):
-    #     embeddings[prot_name] = torch.from_numpy(h5fi[prot_name][:, :])
-    embeddings = load_hdf5_parallel(embedding_h5, all_proteins)
+
+    embeddings = {}
+    with h5py.File(embedding_h5, "r") as h5fi:
+        for prot_name in tqdm(all_proteins):
+            embeddings[prot_name] = torch.from_numpy(h5fi[prot_name][:, :])
+    # embeddings = load_hdf5_parallel(embedding_h5, all_proteins)
 
     # Topsy-Turvy
     run_tt = args.run_tt
@@ -520,10 +630,18 @@ def train_model(args, output):
         glider_mat, glider_map = (None, None)
 
     if args.checkpoint is None:
-
         # Create embedding model
         input_dim = args.input_dim
+
+        ############### foldseek code ###########################
+
+        if allow_foldseek and add_first:
+            input_dim += len(fold_vocab)
+
+        ##########################################################
+
         projection_dim = args.projection_dim
+
         dropout_p = args.dropout_p
         embedding_model = FullyConnectedEmbed(
             input_dim, projection_dim, dropout=dropout_p
@@ -539,7 +657,10 @@ def train_model(args, output):
         log(f"\thidden_dim: {hidden_dim}", file=output)
         log(f"\tkernel_width: {kernel_width}", file=output)
 
-        contact_model = ContactCNN(projection_dim, hidden_dim, kernel_width)
+        proj_dim = projection_dim
+        if allow_foldseek and not add_first:
+            proj_dim += len(fold_vocab)
+        contact_model = ContactCNN(proj_dim, hidden_dim, kernel_width)
 
         # Create the full model
         do_w = not args.no_w
@@ -565,7 +686,7 @@ def train_model(args, output):
 
     else:
         log(
-            "Loading model from checkpoint {}".format(args.checkpoint),
+            f"Loading model from checkpoint {args.checkpoint}",
             file=output,
         )
         model = torch.load(args.checkpoint)
@@ -595,14 +716,11 @@ def train_model(args, output):
     log(f"\tcontact map weight: {cmap_weight}", file=output)
     output.flush()
 
-    batch_report_fmt = (
-        "[{}/{}] training {:.1%}: Loss={:.6}, Accuracy={:.3%}, MSE={:.6}"
-    )
+    batch_report_fmt = "[{}/{}] training {:.1%}: Loss={:.6}, Accuracy={:.3%}, MSE={:.6}"
     epoch_report_fmt = "Finished Epoch {}/{}: Loss={:.6}, Accuracy={:.3%}, MSE={:.6}, Precision={:.6}, Recall={:.6}, F1={:.6}, AUPR={:.6}"
 
     N = len(train_iterator) * batch_size
     for epoch in range(num_epochs):
-
         model.train()
 
         n = 0
@@ -611,8 +729,7 @@ def train_model(args, output):
         mse_accum = 0
 
         # Train batches
-        for (z0, z1, y) in train_iterator:
-
+        for z0, z1, y in train_iterator:
             loss, correct, mse, b = interaction_grad(
                 model,
                 z0,
@@ -625,6 +742,10 @@ def train_model(args, output):
                 glider_map=glider_map,
                 glider_mat=glider_mat,
                 use_cuda=use_cuda,
+                allow_foldseek=allow_foldseek,
+                fold_record=fold_record,
+                fold_vocab=fold_vocab,
+                add_first=add_first,
             )
 
             n += b
@@ -658,7 +779,6 @@ def train_model(args, output):
         model.eval()
 
         with torch.no_grad():
-
             (
                 inter_loss,
                 inter_correct,
@@ -667,7 +787,16 @@ def train_model(args, output):
                 inter_re,
                 inter_f1,
                 inter_aupr,
-            ) = interaction_eval(model, test_iterator, embeddings, use_cuda)
+            ) = interaction_eval(
+                model,
+                test_iterator,
+                embeddings,
+                use_cuda,
+                allow_foldseek,
+                fold_record,
+                fold_vocab,
+                add_first,
+            )
             tokens = [
                 epoch + 1,
                 num_epochs,
@@ -684,12 +813,7 @@ def train_model(args, output):
 
             # Save the model
             if save_prefix is not None:
-                save_path = (
-                    save_prefix
-                    + "_epoch"
-                    + str(epoch + 1).zfill(digits)
-                    + ".sav"
-                )
+                save_path = save_prefix + "_epoch" + str(epoch + 1).zfill(digits) + ".sav"
                 log(f"Saving model to {save_path}", file=output)
                 model.cpu()
                 torch.save(model, save_path)
@@ -721,7 +845,7 @@ def main(args):
         output = open(output, "w")
 
     log(f"D-SCRIPT Version {__version__}", file=output, print_also=True)
-    log(f'Called as: {" ".join(sys.argv)}', file=output, print_also=True)
+    log(f"Called as: {' '.join(sys.argv)}", file=output, print_also=True)
 
     # Set the device
     device = args.device
